#!/usr/bin/perl -sw
##
##
##
## Copyright (c) 2001, Vipul Ved Prakash.  All rights reserved.
## This code is free software; you can redistribute it and/or modify
## it under the same terms as Perl itself.
##
## $Id: Version.pm,v 1.45 2007/05/10 20:32:10 rsoderberg Exp $

package Razor2::Client::Version;
use strict;

our $PROTOCOL = 3;
our $VERSION  = '2.85';

1;

